Attribute VB_Name = "ConvertModule"
Option Explicit

Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" _
    (ByRef hpvDest As Any, ByRef hpvSource As Any, ByVal cbCopy As Long)
Private Declare Sub FillMemory Lib "kernel32" Alias "RtlFillMemory" _
    (ByRef hpvDest As Any, ByVal cbCopy As Long, ByVal Char As Byte)

Private Type TGF_BLOCK
    Pos As Long
    Type As Integer
    Length As Long
End Type

Private MsgLabel As Label

' Note for the Convert___Block-Functions:
'   Argument 1 - CurPos: taken as ByRef (so no additional calculation is needed)
'   Argument 2 - DstPos: taken as ByVal (and is left untouched)
'   Return Value: Number of Bytes Written
' The new File Offset would be: NewPos + ReturnValue

Public Function ConvertKnP2TGF(ByVal GameFileTitle As String, _
                                ByVal Flags As Byte) As Byte

    ' Flags:
    '   Bit 0 (01): light conversion (skips mus/snd files)
    
    Dim CurPos As Long
    Dim NewPos As Long
    Dim LvlCount As Long
    Dim LvlTOC() As Long
    Dim CurLvl As Integer
    'Dim TempByt As Byte
    Dim TempSht As Integer
    Dim TempLng As Long
    Dim TempStr As String
    Dim FileBlks() As TGF_BLOCK ' global blocks 000A - 000D
    Dim CurBlk As TGF_BLOCK
    Dim TempArr() As Byte
    
    ReDim FileBlks(&H0 To &H3)  ' 00 - img, 01 - mtf, 02 - mus, 03 - snd
    
    ' Check version number
    Call CopyMemory(TempSht, FileData(&H4), &H2)
    If (TempSht And &HFF00) <> &H100 Then GoTo ReturnError
    
    NewSize = FileSize * 1.5
    For CurLvl = &H0 To &H3
        If CurLvl >= &H2 And CBool(Flags And &H1) Then Exit For
        Select Case CurLvl
        Case &H0
            TempStr = "img"
        Case &H1
            TempStr = "mtf"
        Case &H2
            TempStr = "mus"
        Case &H3
            TempStr = "snd"
        End Select
        If Dir(GameFileTitle & "." & TempStr) <> "" Then
            FileBlks(CurLvl).Type = &HA + CurLvl
            FileBlks(CurLvl).Length = FileLen(GameFileTitle & "." & TempStr)
            NewSize = NewSize + FileBlks(CurLvl).Length
        End If
    Next CurLvl
    
    ReDim NewData(&H0 To NewSize - 1)
    
    Call CopyMemory(NewData(&H0), FileData(&H0), &H166)
    ' Copy header and write the new version
    TempSht = &H207
    Call CopyMemory(NewData(&H4), TempSht, &H2)
    NewData(&H101) = NewData(&H101) Or &H80
    
    ' Read Level Table
    Call CopyMemory(LvlCount, FileData(&H15C), &H2)
    ReDim LvlTOC(&H0 To LvlCount - 1)
    TempLng = LvlCount * &H4&
    Call CopyMemory(LvlTOC(&H0), FileData(&H166), TempLng)
    
    CurPos = &H166 + &H400
    NewPos = &H166 + TempLng
    
    For CurLvl = &H0 To &H2
        Select Case CurLvl
        Case &H0    ' Block 0007 (Level IDs)
            CurBlk.Type = &H7
            TempSht = &H200
            CurBlk.Length = LvlCount * &H2&
        Case &H1    ' Block 0001 (4-Bit Game Icon)
            Call CopyMemory(CurPos, FileData(&H162), &H4)
            CurBlk.Type = &H1
            TempSht = &H280
            CurBlk.Length = &H340
        Case &H2    ' Block 0002 (Game Preview Bitmap)
            Call CopyMemory(CurPos, FileData(&H158), &H4)
            CurBlk.Type = &H2
            TempSht = &H3428
            CurBlk.Length = &H3428
        End Select
        If TempSht < CurBlk.Length Then
            TempLng = TempSht
        Else
            TempLng = CurBlk.Length
        End If
        
        Call CopyMemory(NewData(NewPos + &H0), CurBlk.Type, &H2)
        Call CopyMemory(NewData(NewPos + &H2), CurBlk.Length, &H4)
        Call CopyMemory(NewData(NewPos + &H6), FileData(CurPos), TempLng)
        CurPos = CurPos + TempSht
        NewPos = NewPos + &H6 + CurBlk.Length
    Next CurLvl
    
    ' Block 0003 (Global Events)
    CurBlk.Type = &H3
    CurBlk.Length = &H0
    Call CopyMemory(NewData(NewPos + &H0), CurBlk.Type, &H2)
    Call CopyMemory(NewData(NewPos + &H2), CurBlk.Length, &H4)
    NewPos = NewPos + &H6 + CurBlk.Length
    
    ' Block 0004 (Extention List)
    CurBlk.Type = &H4
    CurBlk.Length = &H0
    Call CopyMemory(NewData(NewPos + &H0), CurBlk.Type, &H2)
    Call CopyMemory(NewData(NewPos + &H2), CurBlk.Length, &H4)
    NewPos = NewPos + &H6 + CurBlk.Length
    
    ' Block 0005 (Game Palette)
    Call CopyMemory(CurPos, FileData(&H158), &H4)
    CurBlk.Type = &H5
    CurBlk.Length = CopyBmp2Pal(CurPos, NewPos + &H6)
    Call CopyMemory(NewData(NewPos + &H0), CurBlk.Type, &H2)
    Call CopyMemory(NewData(NewPos + &H2), CurBlk.Length, &H4)
    NewPos = NewPos + &H6 + CurBlk.Length
    
    ' Block 0009 (File Menu)
    CurBlk.Type = &H9
    CurBlk.Length = &H0
    Call CopyMemory(NewData(NewPos + &H0), CurBlk.Type, &H2)
    Call CopyMemory(NewData(NewPos + &H2), CurBlk.Length, &H4)
    NewPos = NewPos + &H6 + CurBlk.Length
    
    ' Block 0006 (Global Objects)
    CurBlk.Type = &H6
    CurBlk.Length = &H0
    Call CopyMemory(NewData(NewPos + &H0), CurBlk.Type, &H2)
    Call CopyMemory(NewData(NewPos + &H2), CurBlk.Length, &H4)
    NewPos = NewPos + &H6 + CurBlk.Length
    
    For CurLvl = &H0 To LvlCount - 1
        If Not (MsgLabel Is Nothing) Then
            MsgLabel.Caption = "Converting Level " & Format$(CurLvl + 1) & _
                                " / " & Format$(LvlCount) & " ..."
            MsgLabel.Refresh
        End If
        CurPos = LvlTOC(CurLvl)
        
        ' Block 0008 (Level)
        CurBlk.Type = &H8
        CurBlk.Length = ConvertLevelBlock(CurPos, NewPos + &H6)
        
        Call CopyMemory(NewData(NewPos + &H0), CurBlk.Type, &H2)
        Call CopyMemory(NewData(NewPos + &H2), CurBlk.Length, &H4)
        LvlTOC(CurLvl) = NewPos + &H6
        NewPos = NewPos + &H6 + CurBlk.Length
    Next CurLvl
    
    ' write Level TOC
    TempLng = LvlCount * &H4&
    Call CopyMemory(NewData(&H162), TempLng, &H4)
    Call CopyMemory(NewData(&H166), LvlTOC(&H0), TempLng)
    
    ' write Position of first Non-Game Data-Segment
    Call CopyMemory(NewData(&H158), NewPos, &H4)
    
    For CurLvl = &H0 To &H3
        If FileBlks(CurLvl).Length > &H0 Then
            CurBlk = FileBlks(CurLvl)
            CurBlk.Pos = NewPos
            
            Select Case CurLvl
            Case &H0
                TempStr = "img"
            Case &H1
                TempStr = "mtf"
            Case &H2
                TempStr = "mus"
            Case &H3
                TempStr = "snd"
            End Select
            Open GameFileTitle & "." & TempStr For Binary Access Read As #3
                CurBlk.Length = LOF(3)
                ReDim TempArr(&H0 To CurBlk.Length - 1)
                Get #3, 1, TempArr()
            Close #3
            
            Call CopyMemory(NewData(NewPos + &H0), CurBlk.Type, &H2)
            Call CopyMemory(NewData(NewPos + &H2), CurBlk.Length, &H4)
            Call CopyMemory(NewData(NewPos + &H6), TempArr(&H0), CurBlk.Length)
            NewPos = NewPos + &H6 + CurBlk.Length
        End If
    Next CurLvl
    
    ' Mark last Block
    CurBlk.Type = &H8000 Or CurBlk.Type
    Call CopyMemory(NewData(CurBlk.Pos + &H0), CurBlk.Type, &H2)
    
    NewSize = NewPos
    ReDim Preserve NewData(&H0 To NewSize - 1)
    
    ConvertKnP2TGF = &H0
    Exit Function

ReturnError:

    ConvertKnP2TGF = &HFF
    Exit Function

End Function

Private Function CopyBmp2Pal(ByVal BmpPos As Long, ByVal PalPos As Long) As Long

    Dim CurPos As Long
    Dim NewPos As Long
    Dim CurCol As Integer
    Dim TempSht As Integer
    Dim TempLng As Long
    
    CurPos = BmpPos
    NewPos = PalPos
    
    Call CopyMemory(TempLng, FileData(CurPos + &H0), &H4)
    CurPos = CurPos + TempLng   ' Skip BMP Header
    ' Note: There's no check, if the Bitmap really HAS a palette.
    
    TempSht = &H300
    Call CopyMemory(NewData(NewPos + &H0), TempSht, &H2)
    TempSht = &H100
    Call CopyMemory(NewData(NewPos + &H2), TempSht, &H2)
    NewPos = NewPos + &H4
    
    For CurCol = &H0 To &HFF
        NewData(NewPos + &H0) = FileData(CurPos + &H2)
        NewData(NewPos + &H1) = FileData(CurPos + &H1)
        NewData(NewPos + &H2) = FileData(CurPos + &H0)
        NewData(NewPos + &H3) = &H0
        CurPos = CurPos + &H4
        NewPos = NewPos + &H4
    Next CurCol
    
    CopyBmp2Pal = NewPos - PalPos

End Function

Private Function ConvertLevelBlock(ByRef CurPos As Long, ByVal DstPos As Long) _
                                    As Long

    Dim NewPos As Long
    'Dim TempByt As Byte
    Dim TempSht As Integer
    Dim TempLng As Long
    Dim CurBlk As TGF_BLOCK
    Dim ObjCount As Integer
    Dim CurObj As Integer
    
    NewPos = DstPos
    
    Call CopyMemory(NewData(NewPos + &H0), FileData(CurPos + &H0), &H76)
    ' TODO: correct additional data
    TempLng = &H0
    Call CopyMemory(NewData(NewPos + &H38), TempLng, &H4)
    Call CopyMemory(TempSht, FileData(CurPos + &H4), &H2)
    TempSht = TempSht / 2
    Call CopyMemory(NewData(NewPos + &H46), TempSht, &H2)
    Call CopyMemory(TempSht, FileData(CurPos + &H6), &H2)
    TempSht = TempSht / 2
    Call CopyMemory(NewData(NewPos + &H48), TempSht, &H2)
    TempLng = &HFFFFFF
    Call CopyMemory(NewData(NewPos + &H4A), TempLng, &H4)
    TempSht = &H0
    Call CopyMemory(NewData(NewPos + &H76), TempSht, &H2)
    Call FillMemory(NewData(NewPos + &H78), &H40, &H0)  ' Fade-Names
    NewPos = NewPos + &HB8
    
    ' Preview Bitmap Position
    Call CopyMemory(TempLng, FileData(CurPos + &H38), &H4)
    ' Block 0000 (Level Palette)
    TempLng = CurPos + TempLng
    CurBlk.Type = &H0
    CurBlk.Length = CopyBmp2Pal(TempLng, NewPos + &H6)
    Call CopyMemory(NewData(NewPos + &H0), CurBlk.Type, &H2)
    Call CopyMemory(NewData(NewPos + &H2), CurBlk.Length, &H4)
    NewPos = NewPos + &H6 + CurBlk.Length
    
    ' Block 0001 (Level Preview Bitmap)
    CurBlk.Type = &H1
    CurBlk.Length = &H1028
    
    Call CopyMemory(NewData(NewPos + &H0), CurBlk.Type, &H2)
    Call CopyMemory(NewData(NewPos + &H2), CurBlk.Length, &H4)
    Call CopyMemory(NewData(NewPos + &H6), FileData(TempLng), CurBlk.Length)
    NewPos = NewPos + &H6 + CurBlk.Length
    
    CurPos = CurPos + &H76
    
    ' Block 0002 (Level Object Definition)
    CurBlk.Type = &H2
    CurBlk.Length = ConvertObjectsBlock(CurPos, NewPos + &H6)
    
    Call CopyMemory(NewData(NewPos + &H0), CurBlk.Type, &H2)
    Call CopyMemory(NewData(NewPos + &H2), CurBlk.Length, &H4)
    NewPos = NewPos + &H6 + CurBlk.Length
    
    ' Block 0003 (Level Object Placement)
    CurBlk.Pos = NewPos
    CurBlk.Type = &H3
    NewPos = NewPos + &H6
    
    Call CopyMemory(ObjCount, FileData(CurPos + &H0), &H2)
    If ObjCount > 8192 Then Stop
    Call CopyMemory(NewData(NewPos + &H0), ObjCount, &H2)
    CurPos = CurPos + &H2
    NewPos = NewPos + &H2
    
    ' Copy Object Pointers
    TempLng = &H4& * ObjCount
    Call CopyMemory(NewData(NewPos), FileData(CurPos), TempLng)
    
    TempSht = &H0
    For CurObj = &H0 To ObjCount - 1
        Call CopyMemory(TempLng, FileData(CurPos), &H4)
        If TempLng >= &H0 Then
            TempSht = TempSht + &H1
        End If
        CurPos = CurPos + &H4
        NewPos = NewPos + &H4
    Next CurObj
    
    ' Copy Object Data
    TempLng = &H20& * TempSht
    Call CopyMemory(NewData(NewPos), FileData(CurPos), TempLng)
    CurPos = CurPos + TempLng
    NewPos = NewPos + TempLng
    
    CurBlk.Length = NewPos - (CurBlk.Pos + &H6)
    Call CopyMemory(NewData(CurBlk.Pos + &H0), CurBlk.Type, &H2)
    Call CopyMemory(NewData(CurBlk.Pos + &H2), CurBlk.Length, &H4)
    
    ' Block 0004 (Level Events)
    CurBlk.Type = &H4 Or &H8000
    CurBlk.Length = ConvertEventBlock(CurPos, NewPos + &H6)
    
    Call CopyMemory(NewData(NewPos + &H0), CurBlk.Type, &H2)
    Call CopyMemory(NewData(NewPos + &H2), CurBlk.Length, &H4)
    NewPos = NewPos + &H6 + CurBlk.Length
    
    ConvertLevelBlock = NewPos - DstPos

End Function

Private Function ConvertObjectsBlock(ByRef CurPos As Long, _
                                    ByVal DstPos As Long) As Long

    Const VALUE_STR As String = "Value "
    Dim NewPos As Long
    'Dim TempByt As Byte
    Dim TempSht As Integer
    Dim TempLng As Long
    Dim CurBlk As TGF_BLOCK
    Dim ObjCount As Integer
    Dim CurObj As Integer
    Dim ObjType As Byte
    Dim ValStr() As Byte
    Dim ObjDataLen As Integer
    Dim ObjDataPos As Long
    
    NewPos = DstPos
    
    ReDim ValStr(&H0 To &H17)
    Call FillMemory(ValStr(&H0), &H18, &H0)
    TempLng = Len(VALUE_STR)
    Call CopyMemory(ValStr(&H2), ByVal VALUE_STR, TempLng)
    
    Call CopyMemory(ObjCount, FileData(CurPos + &H0), &H2)
    ' Copy Object Count and an unknown value
    Call CopyMemory(NewData(NewPos), FileData(CurPos), &H4)
    CurPos = CurPos + &H6   ' KnP files store another unknown value
    NewPos = NewPos + &H4
    
    For CurObj = &H0 To ObjCount - 1
        ' Object Data (Main)
        ObjType = FileData(CurPos + &H2A)   ' Object Type
        If ObjType < &HFF Then
            Call CopyMemory(ObjDataLen, FileData(CurPos + &H3E), &H2)
        Else
            ObjDataLen = &H0
        End If
        Call CopyMemory(NewData(NewPos), FileData(CurPos), &H3E)
        ' NewPos + &H40 is written later
        CurPos = CurPos + &H3E
        NewPos = NewPos + &H44
        Call CopyMemory(NewData(NewPos - &H4), NewPos, &H4)
        
        ObjDataPos = CurPos
        CurBlk.Pos = NewPos
        CurBlk.Type = &H0
        
        If ObjType >= &H2 And ObjType < &HFF Then
            ' Object Data (Sub, all objects but Backdrop and Deleted)
            CurPos = CurPos + &H6
            NewPos = NewPos + &H6
            
            ' copy some bytes (I'm doing any fixes just below)
            Call CopyMemory(NewData(NewPos + &H4), FileData(CurPos), &HC)
            
            TempSht = &H0   ' Checksum??
            Call CopyMemory(NewData(NewPos + &H0), TempSht, &H2)
            TempSht = &H30
            Call CopyMemory(NewData(NewPos + &H2), TempSht, &H2)
            TempSht = &H4C  ' some sort of address ??
            Call CopyMemory(NewData(NewPos + &H4), TempSht, &H2)
            TempSht = &H217 ' Version
            Call CopyMemory(NewData(NewPos + &H6), TempSht, &H2)
            CurPos = CurPos + &HC
            
            Select Case ObjType
            Case &H2    ' Active Object
                TempSht = &H33C
                Call CopyMemory(NewData(NewPos + &H10), TempSht, &H2)
                Call FillMemory(NewData(NewPos + &H12), &H12, &H0)
                ' NewPos + &H24 is written later
                TempSht = &HFB
                Call CopyMemory(NewData(NewPos + &H28), TempSht, &H2)
                TempLng = &H49525053    ' "SPRI"
                Call CopyMemory(NewData(NewPos + &H2A), TempLng, &H4)
                NewPos = NewPos + &H2E
                
                ' Sprite Data (Movement Data)
                Call CopyMemory(TempSht, FileData(CurPos + &H0), &H2)
                If TempSht > 4096 Then Stop
                Call CopyMemory(NewData(NewPos), FileData(CurPos), TempSht)
                CurPos = CurPos + TempSht
                NewPos = NewPos + TempSht
                
                ' Sprite Data (Animation Data)
                Call CopyMemory(TempSht, FileData(CurPos + &H0), &H2)
                If TempSht > 4096 Then Stop
                Call CopyMemory(NewData(NewPos), FileData(CurPos), TempSht)
                CurPos = CurPos + TempSht
                NewPos = NewPos + TempSht
                
                ' Rewrite Value-Names Pointer
                TempLng = NewPos - (CurBlk.Pos + &H6) + &H2
                Call CopyMemory(NewData(CurBlk.Pos + &H6 + &H24), TempLng, &H4)
                ' Sprite Data (Value-Names)
                TempSht = &H50
                Call CopyMemory(NewData(NewPos + &H0), TempSht, &H2)
                Call FillMemory(NewData(NewPos + &H2), &H6, &H0)
                NewPos = NewPos + &H8
                
                TempLng = Len(VALUE_STR)
                For TempSht = &H0 To &H2
                    ValStr(&H2 + TempLng) = &H41 + TempSht
                    Call CopyMemory(NewData(NewPos), ValStr(&H0), &H18)
                    NewPos = NewPos + &H18
                Next TempSht
            Case &H3    ' Text
                TempSht = &HF20C
                Call CopyMemory(NewData(NewPos + &H10), TempSht, &H2)
                Call FillMemory(NewData(NewPos + &H12), &H16, &H0)
                TempSht = &H65
                Call CopyMemory(NewData(NewPos + &H28), TempSht, &H2)
                TempLng = &H54584554    ' "TEXT"
                Call CopyMemory(NewData(NewPos + &H2A), TempLng, &H4)
                NewPos = NewPos + &H2E
                
                ' Text Data
                Call CopyMemory(TempSht, FileData(CurPos + &H0), &H2)
                If TempSht > 4096 Then Stop
                Call CopyMemory(NewData(NewPos), FileData(CurPos), TempSht)
                CurPos = CurPos + TempSht
                NewPos = NewPos + TempSht
            Case &H4    ' Question
                TempSht = &HAAAA
                Call CopyMemory(NewData(NewPos + &H10), TempSht, &H2)
                Call FillMemory(NewData(NewPos + &H12), &H16, &H0)
                TempSht = &HAA
                Call CopyMemory(NewData(NewPos + &H28), TempSht, &H2)
                TempLng = &H4E545351    ' "QSTN"
                Call CopyMemory(NewData(NewPos + &H2A), TempLng, &H4)
                NewPos = NewPos + &H2E
                
                ' Question Data
                Call CopyMemory(TempSht, FileData(CurPos + &H0), &H2)
                If TempSht > 4096 Then Stop
                Call CopyMemory(NewData(NewPos), FileData(CurPos), TempSht)
                CurPos = CurPos + TempSht
                NewPos = NewPos + TempSht
                Stop
            Case &H5    ' Score
                TempSht = &HFA0C
                Call CopyMemory(NewData(NewPos + &H10), TempSht, &H2)
                Call FillMemory(NewData(NewPos + &H12), &H16, &H0)
                TempSht = &H77
                Call CopyMemory(NewData(NewPos + &H28), TempSht, &H2)
                TempLng = &H45524353    ' "SCRE"
                Call CopyMemory(NewData(NewPos + &H2A), TempLng, &H4)
                NewPos = NewPos + &H2E
                
                ' Score Data #1
                Call CopyMemory(TempSht, FileData(CurPos + &H0), &H2)
                If TempSht > 4096 Then Stop
                Call CopyMemory(NewData(NewPos), FileData(CurPos), TempSht)
                CurPos = CurPos + TempSht
                NewPos = NewPos + TempSht
                
                If CurPos < ObjDataPos + ObjDataLen Then
                    ' Score Data #2
                    Call CopyMemory(TempSht, FileData(CurPos + &H0), &H2)
                    If TempSht > 4096 Then Stop
                    Call CopyMemory(NewData(NewPos), FileData(CurPos), TempSht)
                    CurPos = CurPos + TempSht
                    NewPos = NewPos + TempSht
                End If
            Case &H6    ' Lives
                TempSht = &HFA0C
                Call CopyMemory(NewData(NewPos + &H10), TempSht, &H2)
                Call FillMemory(NewData(NewPos + &H12), &H16, &H0)
                TempSht = &H77
                Call CopyMemory(NewData(NewPos + &H28), TempSht, &H2)
                TempLng = &H4556494C    ' "LIVE"
                Call CopyMemory(NewData(NewPos + &H2A), TempLng, &H4)
                NewPos = NewPos + &H2E
                
                ' Lives Data #1
                Call CopyMemory(TempSht, FileData(CurPos + &H0), &H2)
                If TempSht > 4096 Then Stop
                Call CopyMemory(NewData(NewPos), FileData(CurPos), TempSht)
                CurPos = CurPos + TempSht
                NewPos = NewPos + TempSht
                
                If CurPos < ObjDataPos + ObjDataLen Then
                    ' Lives Data #2
                    Call CopyMemory(TempSht, FileData(CurPos + &H0), &H2)
                    If TempSht > 4096 Then Stop
                    Call CopyMemory(NewData(NewPos), FileData(CurPos), TempSht)
                    CurPos = CurPos + TempSht
                    NewPos = NewPos + TempSht
                End If
            Case &H7    ' Counter
                TempSht = &HFA0C
                Call CopyMemory(NewData(NewPos + &H10), TempSht, &H2)
                Call FillMemory(NewData(NewPos + &H12), &H16, &H0)
                TempSht = &H77
                Call CopyMemory(NewData(NewPos + &H28), TempSht, &H2)
                TempLng = &H52544E43    ' "CNTR"
                Call CopyMemory(NewData(NewPos + &H2A), TempLng, &H4)
                NewPos = NewPos + &H2E
                
                ' Counter Attributes
                Call CopyMemory(TempSht, FileData(CurPos + &H0), &H2)
                If TempSht > 4096 Then Stop
                Call CopyMemory(NewData(NewPos), FileData(CurPos), TempSht)
                CurPos = CurPos + TempSht
                NewPos = NewPos + TempSht
                
                ' Counter Data
                Call CopyMemory(TempSht, FileData(CurPos + &H0), &H2)
                If TempSht > 4096 Then Stop
                Call CopyMemory(NewData(NewPos), FileData(CurPos), TempSht)
                CurPos = CurPos + TempSht
                NewPos = NewPos + TempSht
                
                ' Counter ???
                Call CopyMemory(TempSht, FileData(CurPos + &H0), &H2)
                If TempSht > 4096 Then Stop
                Call CopyMemory(NewData(NewPos), FileData(CurPos), TempSht)
                CurPos = CurPos + TempSht
                NewPos = NewPos + TempSht
            Case Else
                Debug.Print Hex(ObjType)
                Stop
            End Select
            
            CurBlk.Length = NewPos - (CurBlk.Pos + &H6)
            Call CopyMemory(NewData(CurBlk.Pos + &H0), CurBlk.Type, &H2)
            Call CopyMemory(NewData(CurBlk.Pos + &H2), CurBlk.Length, &H4)
            
            TempLng = NewPos - (CurBlk.Pos - &HC)
            Call CopyMemory(NewData(CurBlk.Pos - &H10), TempLng, &H4)
        End If
        CurPos = ObjDataPos + ObjDataLen
        
        If ObjType < &HFF Then
            ' Object Icon
            ' Block 0002 (Object Icon)
            CurBlk.Pos = NewPos
            CurBlk.Type = &H2 Or &H8000 ' it's the last Block for this object
            NewPos = NewPos + &H6
            
            Call CopyMemory(TempLng, FileData(CurPos + &H6), &H4)
            If TempLng > 1536 Then Stop
            ' Icon Header
            Call CopyMemory(NewData(NewPos), FileData(CurPos), &H18)
            CurPos = CurPos + &H18
            NewPos = NewPos + &H18
            ' Icon Data
            Call CopyMemory(NewData(NewPos), FileData(CurPos), TempLng)
            CurPos = CurPos + TempLng
            NewPos = NewPos + TempLng
            
            CurBlk.Length = NewPos - (CurBlk.Pos + &H6)
            Call CopyMemory(NewData(CurBlk.Pos + &H0), CurBlk.Type, &H2)
            Call CopyMemory(NewData(CurBlk.Pos + &H2), CurBlk.Length, &H4)
        End If
    Next CurObj
    
    ConvertObjectsBlock = NewPos - DstPos

End Function

Private Function ConvertEventBlock(ByRef CurPos As Long, ByVal DstPos As Long) _
                                    As Long

    Dim NewPos As Long
    'Dim TempByt As Byte
    Dim TempSht As Integer
    Dim TempLng As Long
    Dim EvtPos As Long
    
    NewPos = DstPos
    
    ' copy "Fra" + 00 Signature
    Call CopyMemory(NewData(NewPos + &H0), FileData(CurPos + &H0), &H4)
    TempSht = &H217 ' Version Number
    Call CopyMemory(NewData(NewPos + &H4), TempSht, &H6)
    CurPos = CurPos + &H6
    NewPos = NewPos + &H6
    
    ' Convert Events (it's quite complete, but still crashes TGF)
    ' TODO: Fix Events (Numbers in Expressions are missing 2 bytes)
    EvtPos = NewPos
    Call CopyMemory(TempSht, FileData(CurPos), &H2) ' Get Event Data Length
    CurPos = CurPos + &H2
    NewPos = NewPos + &H4   ' TGF supports longer Event sections
    
    TempLng = &H0
    Do
        Call CopyMemory(TempSht, FileData(CurPos + &H0), &H2)
        If TempSht < &H0 Then
            TempSht = TempSht - &H4 ' add 4 bytes to the size
        End If
        Call CopyMemory(NewData(NewPos + &H0), TempSht, &H2)
        
        If TempSht < &H0 Then
            TempSht = -TempSht
            Call CopyMemory(NewData(NewPos + &H2), FileData(CurPos + &H2), &H8)
            Call FillMemory(NewData(NewPos + &HA), &H4, &H0)
            Call CopyMemory(NewData(NewPos + &HE), FileData(CurPos + &HA), _
                            TempSht - &HE)
            CurPos = CurPos + (TempSht - &H4)
            NewPos = NewPos + TempSht
        Else
            CurPos = CurPos + &H2
            NewPos = NewPos + &H2
        End If
    Loop Until TempSht = &H0
    TempLng = NewPos - EvtPos
    Call CopyMemory(NewData(EvtPos), TempLng, &H4)
    
    ' small hack to disable events
    TempSht = &H0
    Call CopyMemory(NewData(EvtPos + &H4), TempSht, &H2)
    
    ' Convert Event Block 2
    Call CopyMemory(TempSht, FileData(CurPos + &H0), &H2)
    TempSht = TempSht - &H2 ' TGF has 2 bytes less (KnP had 00 there anyway)
    Call CopyMemory(NewData(NewPos + &H0), TempSht, &H2)
    Call CopyMemory(NewData(NewPos + &H2), FileData(CurPos + &H2), TempSht)
    CurPos = CurPos + (TempSht + &H2)
    NewPos = NewPos + TempSht
    
    ' Block 03 - not yet finished, so just copy and add 8 zero-bytes
    ' veeery cheap, but enough for now
    Call CopyMemory(TempSht, FileData(CurPos + &H0), &H2)
    TempLng = &H2 + TempSht * &H6
    If CurPos + TempLng >= FileSize Then TempLng = FileSize - CurPos
    
    Call CopyMemory(NewData(NewPos), FileData(CurPos), TempLng)
    CurPos = CurPos + TempLng
    NewPos = NewPos + TempLng + &H8
    
    ConvertEventBlock = NewPos - DstPos

End Function

Public Sub Convert_MsgLabel(ByVal DisplayLabel As Label)

    Set MsgLabel = DisplayLabel

End Sub
